;;;(value :q)


(defun rewrite-with-lemma (term lemma ; &extra formals

                                type-alist obj geneqv wrld state
                                fnstack ancestors
                                simplify-clause-pot-lst rcnst gstack
                                ttree)

; The three values returned by this function are: t or nil indicating
; whether lemma was used to rewrite term, the rewritten version of
; term, and the final version of ttree.  This is a No-Change Loser.

  (let ((gstack (push-gframe 'rewrite-with-lemma nil term lemma)))
    (cond ((f-big-clock-negative-p state)
           (mv nil term ttree))
          ((eq (access rewrite-rule lemma :subclass) 'meta)
           (cond
            ((geneqv-refinementp (access rewrite-rule lemma :equiv)
                                 geneqv
                                 wrld)

; We assume that the meta function has defun-mode :logic.  How could it
; be :program if we proved it correct?

             (mv-let
              (erp val latches)
              (ev-fncall-meta (access rewrite-rule lemma :lhs)
                              (list term)
                              (f-decrement-big-clock state))
              (declare (ignore latches))
              (cond
               (erp
                (mv nil term ttree))
               ((equal term val)
                (mv nil term ttree))
               ((termp val wrld)
                (let ((hyp-fn (access rewrite-rule lemma :hyps)))
                  (mv-let
                   (erp evaled-hyp latches)
                   (if (eq hyp-fn nil)
                       (mv nil *t* nil)
                     (ev-fncall-meta hyp-fn
                                     (list term)
                                     (f-decrement-big-clock state)))
                   (declare (ignore latches))
                   (cond
                    (erp
                     (mv nil term ttree))
                    ((termp evaled-hyp wrld)
                     (mv-let
                      (relieve-hyps-ans failure-reason unify-subst ttree1)
                      (rewrite-entry (relieve-hyps

; The next argument of relieve-hyps is a rune on which to "blame" a
; possible force.  We could blame such a force on a lot of things, but
; we'll blame it on the metarule and the term that it's applied to.

                                      (access rewrite-rule lemma :rune)
                                      term
                                      (flatten-ands-in-lit evaled-hyp)

; The meta function has rewritten term to val and has generated a
; hypothesis called evaled-hyp.  Now ignore the metafunction and just
; imagine that we have a rewrite rule (implies evaled-hyp (equiv term
; val)).  The unifying substitution just maps the vars of term to
; themselves.  There may be additional vars in both evaled-hyp and in
; val.  But they are free at the time we do this relieve-hyps.

;;; !!!!!! bug in Version 2.5:        (let ((vars (all-vars val)))
;;; Fix:
                                      (let ((vars (all-vars term)))
                                        (pairlis$ vars vars)))
                                     :obj nil
                                     :geneqv nil)

; If relieve hyps succeeds we get back a unifying substitution that extends
; the identity substitution above.  This substitution might bind free vars
; in the evaled-hyp.

; Why are we ignoring failure-reason?  Do we need to be calling one of the
; brkpt functions?  No, because we don't break on meta rules.  But perhaps we
; should consider allowing breaks on meta rules.

                      (declare (ignore failure-reason))
                      (cond
                       (relieve-hyps-ans
                        (mv-let
                         (rewritten-rhs ttree)
                         (rewrite-entry (rewrite

; Note: The sublis-var below normalizes the explicit constant
; constructors in val, e.g., (cons '1 '2) becomes '(1 . 2).

                                         (sublis-var nil val)

; At one point we ignored the unify-subst constructed and used a
; nil here.  That was unsound if val involved free vars bound by the
; relief of the evaled-hyp.  We must rewrite val under the extended
; substitution.  Often that is just the identity substitution.

                                         unify-subst
                                         'meta)
                                        :ttree

; Should we be pushing executable counterparts into ttrees when we applying
; metafunctions on behalf of meta rules?  NO:  We should only do that if the
; meta-rule's use is sensitive to whether or not they're enabled, and it's not
; -- all that matters is if the rule itself is enabled.

                                        (push-lemma
                                         (geneqv-refinementp
                                          (access rewrite-rule lemma :equiv)
                                          geneqv
                                          wrld)
                                         (push-lemma
                                          (access rewrite-rule lemma :rune)
                                          ttree1)))
                         (mv t rewritten-rhs ttree)))
                       (t (mv nil term ttree)))))
                    (t (mv (er hard 'rewrite-with-lemma
                               "The hypothesis function ~x0 produced the ~
                                non-termp ~x1 on the input term ~x2.  Our ~
                                implementation requires that ~x0 produce a ~
                                term."
                               hyp-fn
                               evaled-hyp
                               term
                               (access rewrite-rule lemma :lhs))
                           term ttree))))))
               (t (mv (er hard 'rewrite-with-lemma
                          "The metafunction ~x0 produced the non-termp ~x1 on ~
                           the input term ~x2. The proof of the correctness of ~
                           ~x0 establishes that the quotations of these two ~
                           s-expressions have the same value, but our ~
                           implementation additionally requires that ~x0 ~
                           produce a term."
                          (access rewrite-rule lemma :lhs)
                          val
                          term)
                      term ttree)))))
            (t (mv nil term ttree))))
          ((not (geneqv-refinementp (access rewrite-rule lemma :equiv)
                                    geneqv
                                    wrld))
           (mv nil term ttree))
          ((eq (access rewrite-rule lemma :subclass) 'definition)
           (mv-let (rewritten-term ttree)
                   (rewrite-entry (rewrite-fncall lemma term))
                   (mv (not (equal term rewritten-term)) rewritten-term ttree)))
          ((and (or (null (access rewrite-rule lemma :hyps))
                    (not (eq obj t))
                    (not (equal (access rewrite-rule lemma :rhs) *nil*)))
                (or (not (being-openedp (ffn-symb term) fnstack))
                    (not (ffnnamep (ffn-symb term)
                                   (access rewrite-rule lemma :rhs)))))
           (let ((lhs (access rewrite-rule lemma :lhs))
                 (rune (access rewrite-rule lemma :rune)))
             (mv-let (unify-ans unify-subst)
                     (one-way-unify-restrictions
                      lhs
                      term
                      (cdr (assoc-equal
                            rune
                            (access rewrite-constant rcnst
                                    :restrictions-alist))))
                     (cond
                      ((and unify-ans
                            (null (brkpt1 lemma term unify-subst type-alist ancestors
                                          ttree
                                          gstack state)))
                       (cond
                        ((null (loop-stopperp
                                (access rewrite-rule lemma :heuristic-info)
                                unify-subst
                                wrld))
                         (prog2$
                          (brkpt2 nil 'loop-stopper
                                  unify-subst gstack nil nil
                                  rcnst state)
                          (mv nil term ttree)))
                        (t
                         (with-accumulated-persistence
                          rune
                          (flg term ttree)
                          (mv-let
                           (relieve-hyps-ans failure-reason unify-subst ttree)
                           (rewrite-entry (relieve-hyps
                                           rune
                                           term
                                           (access rewrite-rule lemma :hyps)
                                           unify-subst)
                                          :obj nil
                                          :geneqv nil)
                           (cond
                            (relieve-hyps-ans
                             (mv-let
                              (rewritten-rhs ttree)
                              (rewrite-entry
                               (rewrite
                                (if (eq unify-ans 'commuted-equalities)
                                    (commute-equalities
                                     (access rewrite-rule lemma :rhs))
                                  (access rewrite-rule lemma :rhs))
                                unify-subst
                                'rhs))
                              (prog2$
                               (brkpt2 t nil unify-subst gstack rewritten-rhs
                                       ttree rcnst state)
                               (mv t rewritten-rhs
                                   (push-lemma (geneqv-refinementp
                                                (access rewrite-rule lemma :equiv)
                                                geneqv
                                                wrld)
                                               (push-lemma
                                                rune
                                                ttree))))))
                            (t (prog2$
                                (brkpt2 nil failure-reason
                                        unify-subst gstack nil nil
                                        rcnst state)
                                (mv nil term ttree)))))))))
                      (t (mv nil term ttree))))))
          (t (mv nil term ttree))))))



(compile 'rewrite-with-lemma)

(lp)
